/*
    This file is part of HPerf.
    Copyright (C) 2020-2024  Laurent Poirrier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include "token.h"



// ************************************************************************
// 
// ************************************************************************
int tokenize(char *s, size_t sz, int max, char **tokens, int *n_r)
{
	size_t o = 0;
	int n = 0;
	int r = -1;
	
	while (1) {
		// line not null-terminated
		if (o >= sz)
			break;
		
		while ((s[o] == ' ') || (s[o] == '\t')
		|| (s[o] == '\n')) {
			o++;
			
			// line not null-terminated
			if (o >= sz)
				break;
		}
		
		// null termination, return success
		if (s[o] == 0) {
			r = 0;
			break;
		}
		
		// too many tokens
		if (n >= max)
			break;
		
		tokens[n] = s + o;

		char c = 0;
		int paren = 0, curly = 0, square = 0, angle = 0;
		
		while (1) {
			c = s[o];
			
			if (c == 0)
				break;
			
			if ((c == ' ') || (c == '\t') || (c == '\n')) {
				if ((paren <= 0) && (curly <= 0)
				&&  (square <= 0) && (angle <= 0))
					break;
			} else if (c == '(') {
				paren++;
			} else if (c == ')') {
				paren--;
			} else if (c == '{') {
				curly++;
			} else if (c == '}') {
				curly--;
			} else if (c == '[') {
				square++;
			} else if (c == ']') {
				square--;
			} else if (c == '<') {
				angle++;
			} else if (c == '>') {
				angle--;
			}
			
			o++;

			// line not null-terminated
			if (o >= sz)
				break;
		}
		
		n++;
		
		if (c == 0) {
			// null termination, return success
			r = 0;
			break;
		}
		
		s[o] = 0;
		o++;
	}
	
	*n_r = n;
	return r;
}


// ************************************************************************
// 
// ************************************************************************
int match(char *s, char ind, char *p, ...)
{
	va_list va;

	va_start(va, p);
	
	size_t so = 0;
	size_t po = 0;
	int r = 0;

	while (1) {
		char pc = p[po]; 
		if (pc != ind) {
			if (s[so] != pc) {
				r = -1;
				break;
			}
			
			if (pc == 0)
				break;
			so++;
			po++;
			continue;
		}
		
		char **memb = va_arg(va, char **);
		
		if (memb)
			*memb = &s[so];
		
		char stop = p[po + 1];
		
		while ((s[so] != stop) && (s[so] != 0))
			so++;
		
		if (s[so] != stop) {
			r = -1;
			break;
		}
		
		s[so] = 0;
		so++;
		
		if (stop == 0)
			break;
		po += 2;
	}
	
	va_end(va);
	
	return r;
}


// ************************************************************************
// 
// ************************************************************************
uint64_t hexparse(char *s, char **stop)
{
	uint64_t r = 0;
	
	while (1) {
		char c = *s;
		
		if ((c >= '0') && (c <= '9')) {
			r = (r << 4) | (c - '0');
		} else if ((c >= 'A') && (c <= 'F')) {
			r = (r << 4) | (c - 'A' + 10);
		} else if ((c >= 'a') && (c <= 'f')) {
			r = (r << 4) | (c - 'a' + 10);
		} else {
			break;
		}
		
		s++;
	}

	if (stop)
		*stop = s;
	
	return r;
}

// ************************************************************************
// 
// ************************************************************************
uint64_t decparse(char *s, char **stop)
{
	uint64_t r = 0;
	
	while (1) {
		char c = *s;
		
		if ((c >= '0') && (c <= '9')) {
			r = (r * 10) + (c - '0');
		} else {
			break;
		}
		
		s++;
	}

	if (stop)
		*stop = s;
	
	return r;
}

// ************************************************************************
// 
// ************************************************************************
int offset_split(char *s, size_t len, uint64_t *offset_r)
{
	size_t o = len;
	size_t digits = 0;
	uint64_t r = 0;
	
	while (1) {
		//  either:        digit
		//	or:  +   0   x
		//          -3  -2  -1  ^ o
		if (o < 3)
			return -1;
		
		char c = s[o - 1];
		
		if ((c >= '0') && (c <= '9')) {
			r |= ((uint64_t)(c - '0')) << (digits * 4);
		} else if ((c >= 'a') && (c <= 'f')) {
			r |= ((uint64_t)(c - 'a' + 10)) << (digits * 4);
		} else if ((c >= 'A') && (c <= 'F')) {
			r |= ((uint64_t)(c - 'A' + 10)) << (digits * 4);
		} else if ((c == 'x') || (c == 'X')) {
			break;
		} else {
			return -1;
		}
		
		digits++;
		o--;
	}
	
	if (digits == 0)
		return -1;
	
	if ((s[o - 3] != '+') || (s[o - 2] != '0'))
		return -1;
	
	s[o - 3] = 0;
	
	if (offset_r)
		*offset_r = r;
	
	return 0;
}
