/*
    This file is part of HPerf.
    Copyright (C) 2020-2024  Laurent Poirrier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#ifndef TOKEN_H
#define TOKEN_H
#include <string.h>
#include <stdint.h>

/*
tokenize():

Tokenize s, a null-terminated string of length at most (sz - 1),
into at most max whitespace-delimited tokens. Upon success, *n_r holds
the number of tokens.

Special treatment is given to:
 - parenthesis ()
 - curly braces {}
 - square brackets []
 - angle brackets <>
Whitespace is ignored (i.e. does not delimit tokens) until, for each
of them, the number of opening brackets matches the number of closing
brackets.
*/
int tokenize(char *s, size_t sz, int max, char **tokens, int *n_r);

/*
match():

Greedily match a pattern while splitting s in-place.

- In the pattern, any non-indicator character must be exactly matched
  (otherwise, -1 is returned), including the null terminator.
- In the pattern, an indicator character greedily matches any substring
  that does not contain the pattern character immediately following the
  indicator. In other words, the greedy matching can only be stopped by
  the character immediately following the indicator (or the end of the
  string s, in which case -1 is returned)
- Matched substring are returned into the var-arg (char *) parameters,
  unless they are NULL.

For example,

  char *first, *second;
  char s[] = "abcWXdefYZ";
  
  match(s, '?', "?WX?YZ", &first, &second);

will:
- replace 'W' and 'Y' by the null character in s[]
- point first to "abc"
- point second to "def"
- return zero
*/
int match(char *s, char indicator, char *pattern, ...);

/*
hexparse() and decparse():

Parse hexadecimal and decimal strings, only digits are allowed
(no whitespace, no + or -, not even the 0x prefix).
*/
uint64_t hexparse(char *s, char **stop);
uint64_t decparse(char *s, char **stop);

/*
offset_split():

If s matches /+0[xX][0-9a-fA-F]$/, then:
- remove the match from s (in-place: '+' replaced by null char)
- store the value (converted from hex) in *offset_r
- return zero
Otherwise, return -1.
*/
int offset_split(char *s, size_t len, uint64_t *offset_r);

#endif
