/*
    This file is part of HPerf.
    Copyright (C) 2020-2024  Laurent Poirrier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#define _GNU_SOURCE
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include "message.h"
#include "pipe.h"
#include "token.h"
#include "prog.h"
#include "trace.h"

// ************************************************************************
// 
// ************************************************************************
struct trace {
	size_t error, read, samples;
};

// ************************************************************************
// 
// ************************************************************************
static void trace_dump(int n, char **token)
{
	for (int i = 0; i < n; i++) {
		ERROR("\t%2d: {%s}\n", i, token[i]);
	}
}

// ************************************************************************
// 
// ************************************************************************
/*
 [0]  1   [2]         3           [4]             5.0
comm pid time: PERF_RECORD_MMAP pid/tid: [0xADDRESS(0xSIZE)
  @ 0|FILE-OFFSET?]: prot /path
 5.1     5.2         [6]    7

 [0]  1   [2]         3            [4]            5.0
comm pid time: PERF_RECORD_MMAP2 pid/tid: [0xADDRESS(0xSIZE)
  @ 0|FILE-OFFSET? maj:min inode inode_gen]: prot /path
 5.1     5.2        [5.3]  [5.4]   [5.5]     [6]    7

*/
static int trace_parse_mmap(struct prog *p, int n, char **token)
{
	if (n < 8)
		return -1;
	
	int v;
	
	if (strcmp(token[3], "PERF_RECORD_MMAP") == 0) {
		v = 1;
	} else if (strcmp(token[3], "PERF_RECORD_MMAP2") == 0) {
		v = 2;
	} else {
		return -2;
	}
	
	char *e;
	
	uint64_t pid = decparse(token[1], &e);
	
	if (*e != 0)
		return -3;
	
	char *start_str, *length_str, *offset_str;
	
	if (v == 1) {
		if (match(token[5], '?', "[0x?(0x?) @ ?]:",
				&start_str, &length_str, &offset_str))
			return -4;
	} else if (v == 2) {
		if (match(token[5], '?', "[0x?(0x?) @ ? ?]:",
				&start_str, &length_str, &offset_str, NULL))
			return -5;
	} else {
		return -6;
	}
	
	uint64_t start = hexparse(start_str, &e);
	
	if (*e != 0)
		return -7;

	uint64_t length = hexparse(length_str, &e);
	
	if (*e != 0)
		return -8;
	
	uint64_t offset;
	
	if ((offset_str[0] == '0') && (offset_str[1] == 'x')) {
		offset = hexparse(offset_str + 2, &e);
	} else {
		offset = decparse(offset_str, &e);
	}
	
	if (*e != 0)
		return -9;
	
	char *path = token[7];
	
	if (pid == 0)
		return 0;
	
	//DEBUG("mmap%d: pid %ld, addr 0x%lx, size 0x%lx, "
	//	"file %s, offset 0x%lx\n",
	//	v, pid, start, length, path, offset);
	
	if (prog_mmap(p, pid, start, length, path, offset))
		return -10;
	
	return 0;
}

// ************************************************************************
/*
 [0]  1   [2]    3       4        5     6...      b
comm pid time: cycles 'cycles:*' ip sym+symoffs (dso)
ip(dso) / ip(dso) / Mis|Predicted  / X|- / A|- / cycles
                       (b+1+k)
*/
static int trace_parse_sample(struct prog *p, int n, char **token)
{
	if (n < 8)
		return -800;
	
	if (strncmp(token[4], "branches:", 9) == 0) {
		// for now, we silently ignore branch events
		return 0;
	}
	
	if (strncmp(token[4], "cycles:", 7) != 0)
		return -4;
	
	char *e;
	
	uint64_t pid = decparse(token[1], &e);
	
	if (*e != 0)
		return -1;
	
	uint64_t addr = hexparse(token[5], &e);
	
	if (*e != 0)
		return -5;
	
	int b = 6;
	char *sym = token[6];
	uint64_t offs = 0;
	
	while (1) {
		if (b + 1 >= n)
			return -6;
		
		if ((b == 6) && (strcmp(token[6], "[unknown]") == 0)) {
			sym = NULL;
			break;
		}
		
		size_t blen = strlen(token[b]);
		
		if (offset_split(token[b], blen, &offs) == 0)
			break;
		
		token[b][blen] = ' ';
		b++;
	}
	
	b++;
	
	char *dso;
	
	if (match(token[b], '?', "(?)", &dso))
		return -7;
	
	//MESSAGE("%s: %s + %s = 0x%lx\n", dso, sym, hoffs, addr);
	
	if (prog_sample(p, pid, addr, dso, sym, offs))
		return -100;
	
	char *pre_dso = NULL;
	uint64_t pre_addr = (uint64_t)-1;
	
	int k = n - 1;
	
	while (k >= b + 1) {
		char *src_as, *src_dso, *dst_as, *dst_dso;
		char *ps, *cs;
		
		// legacy format
		if (match(token[k], '?', "0x?(?)/0x?(?)/?/?/?/?",
				&src_as, &src_dso, &dst_as, &dst_dso,
				&ps, NULL, NULL, &cs) == 0) {
			k--;
		} else {
			if (k - 2 < 8)
				return -1000 - k;
			
			if (match(token[k - 2], '?', "0x?", &src_as))
				return -2000 - k;
			
			if (match(token[k - 1], '?', "(?)/0x?",
					&src_dso, &dst_as))
				return -3000 - k;
			
			if (match(token[k - 0], '?', "(?)/?/?/?/?",
					&dst_dso, &ps, NULL, NULL, &cs))
				return -4000 - k;
			
			k -= 3;
		}
		
		uint64_t src_addr = hexparse(src_as, &e);
		
		if (*e != 0)
			return -5000 - k;
		
		uint64_t dst_addr = hexparse(dst_as, &e);
		
		if (*e != 0)
			return -6000 - k;
		
		int miss;
		
		if ((ps[0] == 'P') && (ps[1] == 0))
			miss = 0;
		else if ((ps[0] == 'M') && (ps[1] == 0))
			miss = 1;
		else
			return -7000 - k;
		
		uint64_t cycles = decparse(cs, &e);
		
		if ((*e != '/') && (*e != 0))
			return -8000 - k;
		
		// so as to not introduce a bias, we discard the first
		// branch from the stack
		if (pre_dso) {
			if (prog_branch(p, pid,
					pre_addr, pre_dso,
					src_addr, src_dso,
					dst_addr, dst_dso,
					miss, cycles))
				return -9000 - k;
		}
		
		pre_dso = dst_dso;
		pre_addr = dst_addr;
	}
	return 0;
}

// ************************************************************************
// 
// ************************************************************************
static int trace_parse_line(struct prog *p, char *buff, size_t size)
{
	// tokenize
	char *token[256];
	int n;
	
	//printf("Parsing {%s}:\n", buff);
	
	if (tokenize(buff, size, 256, token, &n)) {
		ERROR("Unable to tokenize:\n");
		trace_dump(n, token);
		return -1;
	}
	
	//for (int i = 0; i < n; i++) {
	//	printf("\t%2d: {%s}\n", i, token[i]);
	//}
	
	// parse sample
	int sample_fail = trace_parse_sample(p, n, token);
	
	if (sample_fail == 0)
		return 0;
	
	// parse mmap
	int mmap_fail = trace_parse_mmap(p, n, token);
	
	if (mmap_fail == 0)
		return 0;
	
	ERROR("Unable to parse (sample %d, mmap %d):\n",
		sample_fail, mmap_fail);
	
	trace_dump(n, token);
	return -1;
}


// ************************************************************************
// 
// ************************************************************************
int trace_load(struct prog *p, char *path)
{
	char *argv[8];
	int k = 0;
	
	argv[k++] = "perf";
	argv[k++] = "script";
	
	if (path) {
		argv[k++] = "-i";
		argv[k++] = path;
	}
	
	argv[k++] = "--show-mmap-events";
	argv[k++] = "-F";
	argv[k++] = "comm,pid,time,period,event,ip,sym,symoff,dso,brstack";
	argv[k++] = NULL;

	struct piped_input pi;
	
	if (pipe_in(&pi, argv, NULL))
		return -1;
	
	MESSAGE("%s:\n", path);
	size_t lines = 0;
	size_t parsed = 0;
	
	char buff[16384];
	int r = 0;
	
	while (1) {
		if ((parsed > 0) && ((parsed & 0xffff) == 0))
			MESSAGE("  [samples: %6zd k]\n", parsed >> 10);
		
		if (fgets(buff, sizeof(buff), pi.file) != buff) {
			if (ferror(pi.file)) {
				ERROR("fgets(): %s\n", strerror(errno));
				r = -1;
			}
			
			break;
		}
		
		lines++;
		
		if (trace_parse_line(p, buff, sizeof(buff) - 1)) {
			ERROR("^ parsed: %zd\n^", parsed);
			for (int i = 0; i < k - 1; i++) {
				ERROR(" %s", argv[i]);
			}
			ERROR("\n");
		} else {
			parsed++;
		}
	}
	
	MESSAGE("  samples: parsed: %9zd, ignored: %9zd\n",
		parsed, lines - parsed);
	MESSAGE("             hits: %9ld,  unspec: %9ld, orphans: %9ld\n",
		p->samples - p->unspec - p->orphans,
		p->unspec, p->orphans);
	MESSAGE(" branches:   hits: %9zd,  unspec: %9zd, orphans: %9zd\n",
		p->branch_samples - p->branch_unspec - p->branch_orphans,
		p->branch_unspec, p->branch_orphans);
	MESSAGE("     insn:         %9zd\n", p->insn);
	
	int pr = pipe_wait(&pi);
	
	if (pr) {
		ERROR("perf script failed with status %d\n", pr);
		r = -1;
	}
	
	return r;
}


// ************************************************************************
// 
// ************************************************************************
